# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateMV2lev(Template):
    'Fits 2-level multivariate Normal response models, allowing for missing data, and a choice of uniform or Wishart priors. Creates custom C code.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'Normal', 'Multivariate response']
    engines = ['CustomC']

    inputs = '''
y = DataMatrix('Responses: ')
L2ID = IDVector('Level 2 ID: ')
for i in range(0, len(y)):
    context['x'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)

storeresid = Boolean('Store level 2 residuals?')
randu = Boolean('Randomise starting value for level 2 residuals?')

lenbeta = 0
for i in range(0, len(y)):
    lenbeta += len(context['x'+str(i+1)])
    context['miss'+y[i]] = ParamVector(monitor=False)
    if storeresid:
        context['u' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False)
    else:
        context['u' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)

n = len(y)
if n == 1:
    tau = ParamScalar()
    sigma = ParamScalar(modelled = False)
    tau_u = ParamScalar()
    sigma_u = ParamScalar(modelled = False)
else:
    omega_e = ParamMatrix(modelled = False)
    omega_e.size = n
    omega_u = ParamMatrix(modelled = False)
    omega_u.size = n
    d_e = ParamMatrix()
    d_e.size = n
    d_u = ParamMatrix()
    d_u.size = n
    priors = Text('Priors: ', ['Uniform', 'Wishart'])
    if priors == 'Wishart':
        R = List('R matrix: ')
        v = Integer('Degrees of Freedom:')
    priors2 = Text('Level 2 Priors: ', ['Uniform', 'Wishart'])
    if priors2 == 'Wishart':
        R2 = List('R matrix: ')
        v2 = Integer('Degrees of Freedom:')

beta = ParamVector()
beta.ncols = lenbeta
deviance = ParamScalar(modelled = False)
'''

    latex = r'''
<% lenbeta = 0 %>
\begin{aligned} 
\left(
\begin{array}{l}
% for i in range(0, len(y)):
\mbox{${y[i]}}_i 
% if i != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right) & \sim \mbox{N}
\left[ \left(
\begin{array}{l}
% for i in range(0, len(y)):
\mu_{${i}i}
% if i != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right), \Omega_e \right] \\ 
% for i in range(0, len(y)):
\mu_{${i}i} & = ${mmulttex2(context['x' + str(i+1)], r'\beta', 'i', lenbeta)} + u_{${i},${L2ID}(i)} \\ 
<% lenbeta += len(context['x' + str(i+1)]) %>
% endfor

\left(
\begin{array}{l}
% for i in range(0, len(y)):
u_{${i},${L2ID}(i)}
% if i != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right) & \sim \mbox{N}
\left[ \left(
\begin{array}{l}
% for i in range(0, len(y)):
0
% if i != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right), \Omega_{u} \right] \\ 
% if priors == 'Wishart':
\Omega_e & \sim \mbox{inverse Wishart}_{${len(y)}}[${v}*R_e,${v}],R_e=
\left[\begin{array}{
% for i in range(0, len(y)):
l
% endfor
}
<% lenR = 0%>
% for i in range(0, len(y)):
% for j in range(0, len(y)):
% if j <= i:
${R.name[lenR]}
<% lenR += 1 %>
% endif
% if j != len(y) - 1:
&
% else:
% if i != len(y) - 1:
\\ 
% endif
% endif
% endfor
% endfor
\end{array}
\right]
% endif
% if priors == 'Uniform':
\Omega_e & \propto 1
% endif
\\ 
% if priors2 == 'Wishart':
\Omega_u & \sim \mbox{inverse Wishart}_{${len(y)}}[${v2}*R_u,${v2}],R_u=
\left[\begin{array}{
% for i in range(0, len(y)):
l
% endfor
}
<% lenR = 0%>
% for i in range(0, len(y)):
% for j in range(0, len(y)):
% if j <= i:
${R2.name[lenR]}
<% lenR += 1 %>
% endif
% if j != len(y) - 1:
&
% else:
% if i != len(y) - 1:
\\ 
% endif
% endif
% endfor
% endfor
\end{array}
\right]
% endif
% if priors2 == 'Uniform':
\Omega_u & \propto 1
% endif
\\ 
% for i in range(0, lenbeta):
\beta_{${i}} & \propto 1 \\ 
% endfor
\end{aligned} 
'''

    customccode = '''
<%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out
%>
    {
<% n = len(y) %>\\
        // Note currently using a uniform prior for variance matrix
        SymMatrix sbe(${n});
        int vwe = length(${y[0]});
        for(int i = 0; i < length(${y[0]}); i++) {
<% lenbeta = 0 %>\\
% for i in range(0, n):
            double e${i} = double(miss${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)}) - double(u${i}[int(${L2ID}[i])]);
<% lenbeta += len(context['x' + str(i + 1)]) %>\\
% endfor

% for i in range(0, n):
% for j in range(i, n):
            sbe(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }
% if priors == 'Uniform':
        if (runstate == 0) {
% for i in range(0, n):
            sbe(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if priors == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for i in range(0, n):
    for j in range(0, i + 1):
        Rmat[i, j] = float(R[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, n):
% for j in range(i, n):
        sbe(${i}, ${j}) += ${str(Rmat[i][j] * float(v))};
% endfor
% endfor
% endif
        matrix_sym_invinplace(sbe);

% if priors == 'Uniform':
        vwe -= ${n + 1};
% endif
% if priors == 'Wishart':
        vwe += ${v};
% endif

        bool fail = false;
        mat_d_e = dwishart(vwe, sbe, fail);
        mat_omega_e = matrix_sym_inverse(mat_d_e, fail);
        
        SymMatrix sbu(${n});
        for(int i = 0; i < length(u0); i++) {
% for i in range(0, n):
% for j in range(i, n):
            sbu(${i}, ${j}) += double(u${i}[i]) * double(u${j}[i]);
% endfor
% endfor
        }
% if priors2 == 'Uniform':
        if (runstate == 0) {
% for i in range(0, n):
            sbu(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if priors2 == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for i in range(0, n):
    for j in range(0, i + 1):
        Rmat[i, j] = float(R2[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, n):
% for j in range(i, n):
        sbu(${i}, ${j}) += ${str(Rmat[i, j] * float(v2))};
% endfor
% endfor
% endif

        matrix_sym_invinplace(sbu);
% if priors2 == 'Uniform':
        int vwu = length(u0) - ${n + 1};
% endif
% if priors2 == 'Wishart':
        int vwu = length(u0) + ${v2};
% endif
        mat_d_u = dwishart(vwu, sbu, fail);
        mat_omega_u = matrix_sym_inverse(mat_d_u, fail);

                // Step for beta
                std::vector<int> betaind(${lenbeta});
<% betacount = 0 %>
% for i in range(0, n):
% for j in range(0, len(context['x' +str(i+1)])):
                betaind[${betacount}] = ${i};
<% betacount += 1 %>
% endfor
% endfor
                std::vector<double*> tmp_x;
% for i in range(0, n):
% for j in range(0, len(context['x' + str(i+1)])):
                tmp_x.push_back((double*)${context['x' + str(i+1)][j]});
% endfor
% endfor
                RectMatrix mat_x(tmp_x, length(${y[0]}));

                std::vector<double*> tmp_ymiss;
% for i in range(0, n):
                tmp_ymiss.push_back(miss${y[i]});
% endfor
                RectMatrix mat_ymiss(tmp_ymiss, length(${y[0]}));

                std::vector<double*> tmp_u;
% for i in range(0, n):
                tmp_u.push_back(u${i});
% endfor
                RectMatrix mat_u(tmp_u, length(u0));

                SymMatrix db(${lenbeta});
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            db(j1, j2) += mat_x(i, j1) * mat_d_e(betaind[j1], betaind[j2]) * mat_x(i, j2);
                        }
                    }
                }

                RectMatrix xtdy(${lenbeta}, 1);
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < ${n}; j2++) {
                            xtdy(j1, 0) += mat_x(i, j1) * mat_d_e(betaind[j1], j2) * (mat_ymiss(i, j2) - mat_u(${L2ID}[i], j2));
                        }
                    }
                }

                fail = false;
                //int err = matrix_sym_invinplace(db);
                //RectMatrix betahat = db * xtdy;
                //mat_beta = dmultnormal(betahat, db, fail);

                RectMatrix betahat = matrix_solve(db, xtdy);
                mat_beta = dmultnormalinv(betahat, db, fail);

                std::vector<int> betamap(${n});
                int pos = 0;
% for i in range(0, n):
                betamap[${i}] = pos;
                pos += mat_x${i + 1}.cols();
% endfor

                for (int i = 0; i < mat_u0.rows(); i++) {
                    SymMatrix duj(${n});
                    for (int j = 0; j < ${n}; j++) {
                        for (int k = 0; k < ${n}; k++) {
                            duj(j, k) = mat_d_u(j, k) + mat_d_e(j, k) * idmap_${L2ID}[i].size();
                        }
                    }
                    //SymMatrix dujinv = matrix_sym_inverse(duj, fail);

                    RectMatrix ymxb(${n}, 1);
                    for (int id_k = 0; id_k < idmap_${L2ID}[i].size(); id_k++) {
                        int k = idmap_${L2ID}[i][id_k];
% for j in range(0, n):
                        ymxb(${j}, 0) += mat_ymiss(k, ${j});
                        for (int m = 0; m < mat_x${j + 1}.cols(); m++) {
                            ymxb(${j}, 0) -= mat_x${j + 1}(k, m) * mat_beta(betamap[${j}] + m, 0);
                        }
% endfor
                    }
                    RectMatrix newuj = dmultnormalinv(matrix_solve(duj, mat_d_e * ymxb), duj, fail);

                    for (int j = 0; j < ${n}; j++) {
% if randu:
                        if (runstate != 0) {
                            mat_u(i, j) = newuj(j, 0);
                        } else {
                            mat_u(i, j) = dnorm(0, 1);
                        }
% else:
                        mat_u(i, j) = newuj(j, 0);
% endif
                    }
                }

% for i in range(0, n):
                for (int j = 0; j < mat_ymiss.rows(); j++) {
                    if (mat_y(j, ${i}) <= -9.999e29) {
                        double tmp = mat_u${i}(${L2ID}[j], 0);
                        for (int k = 0; k < mat_x${i + 1}.cols(); k++) {
                            tmp += mat_beta(betamap[${i}] + k, 0) * mat_x${i + 1}(j, k);
                        }
                        for (int k = 0; k < mat_ymiss.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_e(${i}, k) * mat_ymiss(j, k) * (1.0 / mat_d_e(${i}, ${i}));
                            }
                        }
% for m in range(0, n):
% if i != m:
                        tmp += mat_d_e(${i}, ${m}) * mat_u${m}(${L2ID}[j], 0) * (1.0 / mat_d_e(${i}, ${i}));
                        for (int k = 0; k < mat_x${m + 1}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1 / mat_d_e(${i}, ${i})) * mat_beta(betamap[${m}] + k, 0) * mat_x${m + 1}(j, k);
                        }
% endif
% endfor
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    }
                }
% endfor
    }

'''

    preparedata = '''
mydata = data['datafile']

for i in range(0, len(y)):
    context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
'''
